package abbot.editor.widgets;

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.tree.*;
import java.lang.ref.*;

import junit.extensions.abbot.*;
import abbot.Log;
import abbot.tester.*;

// FIXME add a test for a JViewport

public class AbstractComponentDecoratorTest extends ComponentTestFixture {

    private static class Decorator extends AbstractComponentDecorator {
        public Decorator(JComponent c) { super(c); }
        public void paint(Graphics g) {
            Log.debug("paint");
            g.setColor(Color.red);
            g.fillRect(0, 0, 100, 100);
        }
    }

    private static class DecoratedLabel extends JLabel {
        public DecoratedLabel(String text) {
            super(text);
        }
    }

    private class MouseWatcher extends MouseAdapter {
        public boolean gotPress = false;
        public void mousePressed(MouseEvent e) {
            gotPress = true;
        }
    }
    public void testNoEffectOnInput() {
        MouseWatcher watcher = new MouseWatcher();
        JLabel label = new DecoratedLabel(getName());
        label.addMouseListener(watcher);
        showFrame(label);
        new Decorator(label);
        getRobot().waitForIdle();
        getRobot().click(label);
        getRobot().waitForIdle();
        assertTrue("Input not passed to label", watcher.gotPress);
    }

    public void testDecorateBeforeShow() {
        JLabel label = new DecoratedLabel(getName());
        new Decorator(label);
        showFrame(label);
        Point where = label.getLocationOnScreen();
        getRobot().waitForIdle();
        assertEquals("Decorator added before show didn't take effect",
                     Color.red, getRobot().sample(where.x, where.y));
    }

    public void testDecorateAfterRepaint() {
        JLabel label = new DecoratedLabel(getName());
        showFrame(label);
        Decorator decorator = new Decorator(label);
        getRobot().waitForIdle();
        Point where = label.getLocationOnScreen();
        assertEquals("Decoration not applied", Color.red,
                     getRobot().sample(where.x, where.y));
        label.repaint();
        getRobot().waitForIdle();
        assertEquals("Decoration not applied after repaint", Color.red,
                     getRobot().sample(where.x, where.y));
        decorator.dispose();
        decorator = null;
        getRobot().waitForIdle();
        assertTrue("Decoration not removed on decorator removal",
                   !Color.red.equals(getRobot().sample(where.x, where.y)));
    }

    public void testDecorateAndUndecorate() {
        JLabel label = new DecoratedLabel(getName());
        showFrame(label);
        Decorator decorator = new Decorator(label);
        getRobot().waitForIdle();
        Point where = label.getLocationOnScreen();
        assertEquals("Decoration not applied", Color.red,
                     getRobot().sample(where.x, where.y));
        WeakReference ref = new WeakReference(decorator);
        decorator.dispose();
        decorator = null;
        getRobot().waitForIdle();
        assertTrue("Decoration not removed",
                   !Color.red.equals(getRobot().sample(where.x, where.y)));
        System.gc();
        assertNull("Reference to decorator still held", ref.get());
    }

    public void testAutoDispose() throws Throwable {
        JLabel label = new DecoratedLabel(getName());
        Frame frame = showFrame(label);
        Decorator decorator = new Decorator(label);
        WeakReference dref = new WeakReference(decorator);
        WeakReference lref = new WeakReference(label);
        decorator = null;
        frame.remove(label);
        assertNull("Label should have no parent", label.getParent());
        label = null;
        // This is required on Linux/1.4.2 to get rid of additional references
        // to the JLabel (focus listeners and such)
        disposeWindow(frame);
        // This should get rid of the label and the decorator
        System.gc();
        assertEquals("Label still referenced", null, lref.get());
        assertEquals("Decorator still referenced", null, dref.get());
    }

    public AbstractComponentDecoratorTest(String name) { super(name); }
    public static void main(String[] args) {
        TestHelper.runTests(args, AbstractComponentDecoratorTest.class);
    }
}
