package abbot.tester;

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.border.*;

import junit.framework.*;
import junit.extensions.abbot.*;
import junit.extensions.abbot.Timer;

public class WindowTrackerTest extends ComponentTestFixture {
    
    private class MouseWatcher extends MouseAdapter {
        private boolean gotClick = false;
        public void mousePressed(MouseEvent e) {
            gotClick = true;
        }
    }

    private WindowTracker tracker;

    private void wait(Window w, boolean state, String msg) {
        Timer timer = new Timer();
        while (tracker.isWindowReady(w) != state) {
            if (timer.elapsed() > WindowTracker.WINDOW_READY_DELAY+500)
                fail(msg);
            try { Thread.sleep(10); } catch(InterruptedException e) { }
        }
    }

    protected void setUp() {
        tracker = new WindowTracker();
    }

    protected void tearDown() {
        tracker = null;
    }

    private class MissedShow1 extends AssertionFailedError {
        public MissedShow1() {
            super("Window not ready after initial show");
        }
    }
    private class MissedShow2 extends AssertionFailedError {
        public MissedShow2() {
            super("Window not ready after subsequent show");
        }
    }

    public void testTrackWindowState() throws Throwable {
        Frame w = new Frame(getName());
        w.pack();
        w.show();
        wait(w, true, "WindowTracker didn't catch initial Window.show()");
        assertTrue("Window should be showing", w.isShowing());

        w.hide();
        wait(w, false, "WindowTracker didn't catch Window.hide()");
        assertTrue("Window should not be showing", !w.isShowing());

        w.show();
        wait(w, true, "WindowTracker didn't catch Window.show() after hide");
        assertTrue("Window should be showing", w.isShowing());

        // Generates WINDOW_CLOSE, may or may not generate COMPONENT_HIDDEN
        w.dispose();
        wait(w, false, "WindowTracker didn't catch Window.dispose()");
        assertTrue("Window should not be showing", !w.isShowing());
    }

    public void testSingleWindowReady() throws Throwable {
        JFrame f = new JFrame(getName());
        MouseWatcher watcher = new MouseWatcher();
        f.addMouseListener(watcher);
        f.getContentPane().add(new JLabel(getName()));
        f.pack();
        f.setSize(100, 100);
        f.show();
        try {
            wait(f, true, "WindowTracker didn't catch initial Window.show");
            getRobot().click(f);
            getRobot().waitForIdle();
            if (!watcher.gotClick)
                throw new MissedShow1();
            
            watcher.gotClick = false;
            f.hide();
            wait(f, false, "WindowTracker didn't catch Window.hide after show");
            
            f.show();
            wait(f, true, "WindowTracker didn't catch Window.show after hide");
            assertTrue("Window not ready", tracker.isWindowReady(f));
            getRobot().click(f);
            getRobot().waitForIdle();
            if (!watcher.gotClick)
                throw new MissedShow2();
        }
        finally {
            f.dispose();
        }
    }

    public void testComplexWindowReady() throws Throwable {
        JList list = new JList(new String[] {
            "one", "two", "three", "four", "five"
        });
        JFrame f = new JFrame(getName() + "1");
        f.getContentPane().add(list);
        MouseWatcher watcher = new MouseWatcher();
        list.addMouseListener(watcher);
        f.pack();
        f.setLocation(100, 100);
        f.show();
        try {
            wait(f, true, "WindowTracker didn't catch window");
            
            getRobot().click(list);
            getRobot().waitForIdle();
            
            assertTrue("Window not ready for click", watcher.gotClick);
        }
        finally {
            f.dispose();
        }
    }

    public void testTrackRepeatedWindowReady() throws Throwable {
        // This test is only required in Robot mode
        if (getRobot().getEventMode() != Robot.EM_ROBOT)
            return;
        int EXPECTED = 100;
        int count1 = EXPECTED;
        int count2 = EXPECTED;
        int count3 = EXPECTED;
        for (int i=0;i < EXPECTED;i++) {
            try {
                testSingleWindowReady();
            }
            catch(MissedShow1 e) {
                --count1;
            }
            catch(MissedShow2 e) {
                --count2;
            }
            try {
                testComplexWindowReady();
            }
            catch(AssertionFailedError e) {
                --count3;
            }
        }
        assertEquals("Missed some clicks on initial show", EXPECTED, count1);
        assertEquals("Missed some clicks on subsequent show", EXPECTED, count2);
        assertEquals("Missed clicks on complex window", EXPECTED, count3);
    }

    public void testTrackWindowReadyAfterDispose() throws Exception {
        Frame w = new Frame(getName());
        w.pack();
        w.show();
        wait(w, true, "Didn't catch initial Window.show()");
        assertTrue("Window not showing", w.isShowing());

        w.dispose();
        wait(w, false, "Didn't catch Window.dispose()");
        assertTrue("Window not hidden", !w.isShowing());
        
        w.pack();
        w.show();
        wait(w, true, "Didn't catch Window.show() after dispose(), pack()");
        assertTrue("Window not showing", w.isShowing());
    }

    public static void main(String[] args) {
        RepeatHelper.runTests(args, WindowTrackerTest.class);
    }
}
